# define context class
# written by MOROHOSHI Akihiko <moro@nii.ac.jp>

require 'tagname'

########################################################################
# Context
class Context
    attr_reader :size, :color
    
    SIZE_UNKNOWN = -1
    COLOR_UNKNOWN = ?U

    MASK_UNDERLINE	= 0b11
    MASK_BOLD		= 4
    MASK_EMPHASIS	= 8
    MASK_STRONG_EMPHASIS= 16
    MASK_ITALIC		= 32
    MASK_TYPEWRITER	= 64
    
    def initialize
	@bits = 0
	@underline = 0
	@size = SIZE_UNKNOWN
	@color = COLOR_UNKNOWN
    end

    attr_reader :bits
    protected :bits

    def underline
	@bits & MASK_UNDERLINE
    end
    def _set_underline(u)
	@bits = ((@bits & ~MASK_UNDERLINE) | u)
    end
    private :_set_underline
    
    ['bold','emphasis','strong_emphasis','italic','typewriter'].each do |name|
	mask = "MASK_"+name.upcase
	module_eval "def #{name}; (@bits & MASK_#{name.upcase}) > 0; end"
	module_eval "def #{name}=(bool); _bit_set(#{mask},bool); end"
    end
    def _bit_set(mask,bool)
	if bool
	    @bits |= mask
	else
	    @bits &= ~mask
	end
    end
    private :_bit_set

    COLORS = [?r, ?g, ?b, ?c, ?m, ?y, ?k, ?w]
    def color=(c)
	if COLORS.include? c
	    @color = c
	else
	    raise "unknown color: #{c.inspect}"
	end
    end

    SIZE_RANGE = (0..9)
    def size=(s)
	if SIZE_RANGE.include? s
	    @size = s
	else
	    raise "invalid size: #{s}"
	end
    end

    def underline_add
	u = underline
	if u < 3
	    _set_underline u+1
	end
    end

    def underline_reset
	_set_underline 0
    end

    def space_context()
	c = self.dup
	c.bold = c.emphasis = c.strong_emphasis = c.italic = false
	c.color = ?w if c.underline == 0
	c
    end

    def ==(other)
	return false if other.type != self.type

	if (@bits & ~MASK_EMPHASIS) != (other.bits & ~MASK_EMPHASIS)
	    false
	elsif @color != other.color
	    false
	elsif @size != other.size
	    false
	elsif self.emphasis != other.emphasis and
		(!self.strong_emphasis or !other.strong_emphasis)
	    false
	else
	    true
	end

	# return false if @bits != other.bits
    end

    def eql?(other)
	self == other
    end

    def hash1
	@bits
    end
    
    def hash
	h = hash1
	h |= (@size << 7)	# 7,8,9,10,11 (5 bit)
	h |= (@color << 12)	# 12-19 (8 bit)
    end

    def distance(other)
	d = 0
	d += 1 if other.bold != self.bold
	d += 1 if other.typewriter != self.typewriter
	d += 1 if other.italic != self.italic
	d += (other.underline - self.underline).abs
	d += 1 if other.color != self.color
	d += 1 if other.size != self.size
	d += 1 if other.strong_emphasis != self.strong_emphasis
	unless (other.strong_emphasis == self.strong_emphasis and
		self.strong_emphasis)
	    d += 1 if other.emphasis != self.emphasis
	end
	
	d
    end

    # $B?dDj:G>.5wN%!#(Best$B$O(BESTimated.
    def distance_min_est(other)
	d = 0
	d1 = 0
	d1 += 1 if other.bold != self.bold
	d1 += 1 if other.typewriter != self.typewriter
	d1 += 1 if other.italic != self.italic
	d1 += 1 if other.strong_emphasis != self.strong_emphasis
	unless (other.strong_emphasis == self.strong_emphasis and
		self.strong_emphasis)
	    d1 += 1 if other.emphasis != self.emphasis
	end
	
	d2 = (other.underline - self.underline).abs
	
	if self.plain? or other.plain?
	    d += d1+d2
	else
	    dd1 = d1+d2
	    plain_context = context_modified_by_tag(self,TAG_PL)
	    dd2 = self.distance_min_est(plain_context) +
		plain_context.distance_min_est(other)
	    d += (dd1 <= dd2 ? dd1 : dd2)
	end

	d += 1 if other.color != @color
	d += 1 if other.size != @size
	
	d
    end

    # $B?dDj:G>.5wN%(B($B%?%0$N%P%$%H?t$K$h$k(B)$B!#(B
    # $B$=$l$>$l$NCM$,(B .5 $BC10L$NC<?t$J$N$O!"(B
    # $B%H!<%/%sNs$N8e$m$+$iE,MQ$7$F$$$/$3$H$rA[Dj$7$F$$$k!#(B
    # $B>o$K3+;O%?%0$h$j$b=*N;%?%0$NJ}$,(B1$BD9$/!"(B
    # $B$+$D!"%H!<%/%sNs$N$"$kE@$h$j8e$mB&$r8+$?>l9g(B
    # $B3+;O%?%0$N8D?t$,=*N;%?%0$N8D?t$h$jB?$$$3$H$O$"$jF@$J$$$N$G(B
    # $BCzEYNI$$!#(B
    def distance_min_est_tag_1(other,pl_both)
	d = 0
	d1 = 0
	d1 += 3.5 if other.bold != self.bold
	d1 += 4.5 if other.typewriter != self.typewriter
	d1 += 3.5 if other.italic != self.italic
	d1 += 3.5 if other.strong_emphasis != self.strong_emphasis
	unless (other.strong_emphasis == self.strong_emphasis and
		self.strong_emphasis)
	    d1 += 4.5 if other.emphasis != self.emphasis
	end
	
	d2 = (other.underline - self.underline).abs * 3.5

	if other.plain?
	    dd1 = d1+d2
	    dd2 = 4.5
	    d += (dd1 <= dd2 ? dd1 : dd2)
	elsif self.plain?
	    dd1 = d1+d2
	    dd2 = 4.5
	    if pl_both
		d += (dd1 <= dd2 ? dd1 : dd2)
	    else
		d += dd1
	    end
	else
	    dd1 = d1+d2
	    plain_context = context_modified_by_tag(self,TAG_PL)
	    dd2 = self.distance_min_est_tag(plain_context) +
		plain_context.distance_min_est_tag(other)
	    d += (dd1 <= dd2 ? dd1 : dd2)
	end

	d
    end
    def distance_min_est_tag(other,pl_both=true)
	d = distance_min_est_tag_1(other,pl_both)
	
	d += 3.5 if other.color != @color
	d += 3.5 if other.size != @size
	
	d
    end

    def plain?
	@is_plain || _plain?
    end

    def _plain?
	@bits == 0
    end

    def freeze
	unless frozen? 
	    @is_plain = _plain?
	    super
	end
	self
    end

    RootContext = Context.new
    RootContext.freeze
end

########################################################################
# utility
def context_modified_by_tag(context,tag)
    cont = context.dup
    case tag
    when TAG_PL
	cont.underline_reset
	cont.bold = false
	cont.emphasis = false
	cont.italic = false
	cont.strong_emphasis = false
	cont.typewriter = false
    when TAG_B
	cont.bold = true
    when TAG_EM
	cont.emphasis = (! cont.emphasis)
    when TAG_S
	cont.strong_emphasis = true
    when TAG_I
	cont.italic = true
    when TAG_TT
	cont.typewriter = true
    when TAG_U
	cont.underline_add
    else
	if s = tag.size		# '0' - '9' 
	    cont.size = s
	elsif c = tag.color	# 'r' - 'w'
	    cont.color = c
	else
	    raise "unknown tag: #{tag}"
	end
    end
    
    cont.freeze
    cont
end
